/*
 * Copyright (c) 2005 Marek Materzok
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

using Nemerle.Compiler;
using Nemerle.Collections;
using PT = Nemerle.Compiler.Parsetree;
using SCG = System.Collections.Generic;

namespace Wraith {

class Exc: System.Exception
{
};

public sealed class StringScanner
{
    public this(s: string) {
        buffer = s;
        position = 0;
    }

    public end(): bool {
        position == buffer.Length
    }

    public Buffer: string {
        get { buffer }
    };

    public Position: int {
        get { position }
        set { position = value }
    }

    public Length: int {
        get { buffer.Length }
    }

    buffer: string;
    mutable position: int;
}

public sealed class TokenScanner['t]
{
    public this(s: array['t]) {
        buffer = s;
        position = 0;
    }

    public end(): bool {
        position == buffer.Length
    }

    public Buffer: array['t] {
        get { buffer }
    }; 

    public Position: int {
        get { position }
        set { position = value }
    }

    public Length: int {
        get { buffer.Length }
    }

    buffer: array['t];
    mutable position: int;
};

public interface IStringParser
{
    Parse(sc: StringScanner): bool;
    Parse(s: string): int;
};

public interface ITokenParser['t]
{
    Parse(sc: TokenScanner['t]): bool;
    Parse(a: array['t]): int;
};

variant CharacterClass
{
    | Whitespace
    | Alphabetic
    | Alphanumeric
    | Numeric
    | Punctuation
    | Symbol
    | Upper
    | Lower
};

variant Expression
{
    | StringMatch {s: string}
    | RegexpMatch {s: string}
    | CharMatch {c: char}
    | CharClass {c: list[CharacterClass]}

    | MatchMatch {p: PT.PExpr}
    
    | Rule {r: string}
    | ChangeSkip {s: string; e: Expression}
    | Epsilon {e: Expression}
    | Optional {e: Expression}
    | Many {e: Expression}
    | Exactly {i: int; e: Expression}
    | AtLeast {i: int; e: Expression}
    | FromTo {i: int; j: int; e: Expression}
    | Sequence {e1: Expression; e2: Expression}
    | Alternative {e1: Expression; e2: Expression}
    | Intersection {e1: Expression; e2: Expression}
    | Code {c: PT.PExpr.Sequence}
    | Nothing {b: bool}
    | Anything
    | EndOfInput
};

class RuleDesc
{
    public this(name: string, expr: Expression, start: PT.PExpr.Sequence, ret: PT.PExpr, rt: MType) {
        this.name = name;
        this.expr = expr;
        this.start = start;
        this.ret = ret;
        this.rt = rt;
        this.nm = Macros.NewSymbol(name);
    }
    public name: string;
    public expr: Expression;
    public start: PT.PExpr.Sequence;
    public ret: PT.PExpr;
    public rt: MType;
    public nm: PT.Name;
};

class ParserDesc
{
    public this(rules: list[RuleDesc], members: list[PT.ClassMember]) {
        this.rules = rules;
        this.members = members;
    }
    
    public rules: list[RuleDesc];
    public members: list[PT.ClassMember];
};

abstract class AbstractParserCompiler
{
    protected ctx: Typer;
    protected t: TypeBuilder;
    protected st_const: MethodBuilder;
    rules: Hashtable[string, RuleDesc] = Hashtable();
    usedr: Hashtable[string, bool] = Hashtable();
    mutable skipper: RuleDesc = null;
    mutable current_rule : RuleDesc = null;

    public this(ctx: Typer, iface: PT.PExpr) {
        this.ctx = ctx;

        def mname = Macros.NewSymbol("parser-type");
        t = ctx.Env.Define( <[ decl: class $(mname : name): $iface {
                static this() {
                }
            } ]> );
        def constl = t.GetConstructors(BindingFlags.Static 
            %| BindingFlags.Public %| BindingFlags.NonPublic 
            %| BindingFlags.DeclaredOnly);
        assert(constl.Length == 1);
        st_const = constl.Head :> MethodBuilder;
    }

    public compile(pdef: ParserDesc): TypeBuilder {
        foreach(member in pdef.members) {
            t.Define(member);
        }
        
        foreach(rule: RuleDesc in pdef.rules) {
            when (rules.ContainsKey(rule.name))
                throw Exc();
            rules.Item[rule.name] = rule;
        }

        when (rules.ContainsKey("skipper"))
            skipper = rules.Item["skipper"];

        foreach(rule: RuleDesc in pdef.rules) {
            t.Define(compile_rule(rule));
        }
        def main = rules.Item["main"];
        when (main == null) throw Exc();
        compile_public(main);
        compile_defaults();
        t.Compile();
        t
    }

    compile_defaults() : void {
        def constl = t.GetConstructors(BindingFlags.Instance 
            %| BindingFlags.Public %| BindingFlags.DeclaredOnly);
        when (constl.IsEmpty) {
            t.Define(<[ decl: public this() {} ]>);
        }
    }

    protected abstract compile_public(main: RuleDesc): void;

    protected abstract build_rule(rule: RuleDesc, exprs: list[PT.PExpr]): PT.ClassMember;

    compile_rule(rule: RuleDesc): PT.ClassMember {
        current_rule = rule;
        usedr.Clear();
        def body = compile_expr(rule.expr);
        def de = <[ 
            if ($body) (true, $(rule.ret))
            else (false, { _s.Position = _OldPosition; $(Macros.DefaultValueOfType(rule.rt)) })
        ]>;
        def varinit = usedr.Fold(rule.start.body + [de],
            fun(n,_,acc) {
                def rinfo2 = rules.Item[n];
                <[ mutable $(rinfo2.name : usesite) = $(Macros.DefaultValueOfType(rinfo2.rt)) ]> :: acc
//                <[ mutable $(PT.Name(rinfo2.desc.name, MacroColorizer.Color, MacroColorizer.UseContext) : name) = $(Macros.DefaultValueOfType(rinfo2.rt)) ]> :: acc
            });
//        Message.Debug(<[ {.. $varinit} ]>.ToString());
        build_rule(rule, varinit);
    }

    protected abstract compile_expr(ex: Expression): PT.PExpr;
    
    protected compile_expr_generic(ex: Expression): PT.PExpr { 
        | Nothing(b) =>
            add_skipper( <[ $(b : bool) ]> );
        | Anything =>
            add_skipper( <[ 
                if (_s.Position < _s.Length) {
                    _s.Position += 1;
                    true
                } else false
            ]> );
        | EndOfInput =>
            add_skipper( <[
                _s.Length == _s.Position
            ]> )
        // Just a optimisation
        | Rule("true") => 
            compile_expr(Expression.Nothing(true));
        | Rule("false") => 
            compile_expr(Expression.Nothing(true));
        | Rule("endofinput") => 
            compile_expr(Expression.EndOfInput());
        | Rule(s) => 
            def ret = Macros.NewSymbol();
            def ts = Macros.NewSymbol();
            def rinfo = rules.Item[s];
            when (rinfo == null) throw Exc();
            usedr.Item[s] = true;
            add_skipper( <[ 
                def ($(ret : name), $(ts : name)) = $(rinfo.nm : name)(_s);
                if ($(ret : name)) {
                    $(s : usesite) = $(ts : name);
//                    $(PT.Name(s, MacroColorizer.Color, MacroColorizer.UseContext) : name) = $(ts : name);
                    true
                } else false
            ]> )
        | ChangeSkip(s, e) =>
            def bskip = skipper;
            if (s != "true") {
                skipper = rules.Item[s];
                when (skipper == null) throw Exc();
            } else skipper = null; // true is like no skipper at all
            def ret = <[ $(compile_expr(e)) ]>;
            skipper = bskip;
            add_skipper(ret)
        | Code(c) =>
            <[
                $c; true
            ]>
        | Epsilon(Code(c)) =>
            <[ $c ]>
        | Epsilon(e) =>
            def mem = Macros.NewSymbol();
            <[
                def $(mem : name) = _s.Position;
                if ($(compile_expr(e))) {
                    _s.Position = $(mem : name);
                    true
                } else false
            ]>
        | Many(e) => 
            def mem = Macros.NewSymbol();
            <[ 
                mutable $(mem : name) = _s.Position;
                while ($(compile_expr(e))) {
                    $(mem : name) = _s.Position; 
                }
                _s.Position = $(mem : name);
                true
            ]>
        | Optional(e) => 
            def mem = Macros.NewSymbol();
            <[ 
                def $(mem : name) = _s.Position;
                if ($(compile_expr(e))) true
                else {
                    _s.Position = $(mem : name);
                    true
                }
            ]>
        | Exactly(i, e) =>
            def mem = Macros.NewSymbol();
            def cnt = Macros.NewSymbol();
            <[
                mutable $(mem : name) = _s.Position;
                mutable $(cnt : name) = 0; 
                while ($(cnt : name) < $(i : int) && $(compile_expr(e))) {
                    $(mem : name) = _s.Position;
                    $(cnt : name)++;
                }
                if ($(cnt : name) == $(i : int)) {
                    _s.Position = $(mem : name);
                    true
                } else false
            ]>
        | AtLeast(i, e) =>
            def mem = Macros.NewSymbol();
            def cnt = Macros.NewSymbol();
            <[
                mutable $(mem : name) = _s.Position;
                mutable $(cnt : name) = 0; 
                while ($(compile_expr(e))) {
                    $(mem : name) = _s.Position;
                    $(cnt : name)++;
                }
                if ($(cnt : name) >= $(i : int)) {
                    _s.Position = $(mem : name);
                    true
                } else false
            ]>
        | FromTo(i, j, e) =>
            def mem = Macros.NewSymbol();
            def cnt = Macros.NewSymbol();
            <[
                mutable $(mem : name) = _s.Position;
                mutable $(cnt : name) = 0; 
                while ($(cnt : name) < $(j : int) && $(compile_expr(e))) {
                    $(mem : name) = _s.Position;
                    $(cnt : name)++;
                }
                if ($(cnt : name) >= $(i : int)) {
                    _s.Position = $(mem : name);
                    true
                } else false
            ]>
        | Sequence(Code(c), e) => 
            <[
                {.. $(c.body + [compile_expr(e)])}
            ]>
        | Sequence(e1, e2) => 
            <[ 
                if ($(compile_expr(e1))) {
                    if ($(compile_expr(e2))) true
                    else false
                } else false
            ]>
        | Alternative(e1, e2) =>
            def mem = Macros.NewSymbol();
            <[ 
                def $(mem : name) = _s.Position;
                if ($(compile_expr(e1))) true
                else {
                    _s.Position = $(mem : name);
                    $(compile_expr(e2))
                }
            ]>
        | Intersection(e1, e2) =>
            def mem = Macros.NewSymbol();
            def mem2 = Macros.NewSymbol();
            <[
                def $(mem : name) = _s.Position;
                if ($(compile_expr(e1))) {
                    def $(mem2 : name) = _s.Position;
                    _s.Position = $(mem : name);
                    if ($(compile_expr(e2))) _s.Position == $(mem2 : name)
                    else false
                } else false
            ]>
        | _ => throw Exc()
    }

    protected add_skipper(e: PT.PExpr): PT.PExpr {
        if (skipper == null || skipper : object == current_rule) e
        else <[
            if ($(skipper.nm : name)(_s)[0]) $e
            else false
        ]>
    }
};

class StringParserCompiler: AbstractParserCompiler 
{
    public this(ctx: Typer) {
        base(ctx, <[ IStringParser ]>);
    }

    protected override compile_public(main: RuleDesc): void {
        t.Define( <[ decl: public ParseGet(s: string): $(main.rt : typed) * int {
                def sc = StringScanner(s);
                def (ok, ret) = $(main.nm : name)(sc);
                if (ok) (ret, sc.Position)
                else (ret, -1)
            } ]> );
        t.Define( <[ decl: public Parse(sc: StringScanner): bool {
                def (ok, _) = $(main.nm : name)(sc);
                ok
            } ]> );
        t.Define( <[ decl: public Parse(s: string): int {
                def sc = StringScanner(s);
                def (ok, _) = $(main.nm : name)(sc);
                if (ok) sc.Position
                else -1
            } ]> );
    }
    
    protected override build_rule(rule: RuleDesc, exprs: list[PT.PExpr]): PT.ClassMember {
        <[ decl: $(rule.nm : name)(_s: StringScanner) : bool * $(rule.rt : typed) {
            def _OldPosition = _s.Position;
            def $("_matched": usesite)() {
                _s.Buffer.Substring(_OldPosition, _s.Position - _OldPosition)
            }
            def $("_matchedfrom": usesite)(pos: int) {
                _s.Buffer.Substring(pos, _s.Position - pos)
            }
            def $("_mark": usesite)() {
                _s.Position
            } 
            {.. $exprs}
        } ]>
    }

    protected override compile_expr(ex: Expression): PT.PExpr { 
        | CharMatch(s) =>
            add_skipper( <[
                if (_s.Buffer.Length == _s.Position) false
                else {
                    if ($(s : char) == _s.Buffer.Chars[_s.Position]) {
                        _s.Position += 1;
                        true
                    } else false
                }
            ]> )
        | CharClass(cc) => 
            def ch = Macros.NewSymbol();
            def cond = cc.FoldLeft(<[ false ]>, fun(x: CharacterClass, ac) {
                    match (x) {
                        | Whitespace =>
                            <[ $ac || char.IsWhiteSpace($(ch : name)) ]>
                        | Alphabetic =>
                            <[ $ac || char.IsLetter($(ch : name)) ]>
                        | Alphanumeric =>
                            <[ $ac || char.IsLetterOrDigit($(ch : name)) ]>
                        | Numeric =>
                            <[ $ac || char.IsDigit($(ch : name)) ]>
                        | Punctuation =>
                            <[ $ac || char.IsPunctuation($(ch : name)) ]>
                        | Symbol =>
                            <[ $ac || char.IsSymbol($(ch : name)) ]>
                        | Lower =>
                            <[ $ac || char.IsLower($(ch : name)) ]>
                        | Upper =>
                            <[ $ac || char.IsUpper($(ch : name)) ]>
                    }
                });
            add_skipper( <[
                if (_s.Buffer.Length == _s.Position) false
                else {
                    def $(ch : name) = _s.Buffer.Chars[_s.Position];
                    if ($cond) {
                        _s.Position += 1;
                        true
                    } else false
                }
            ]> )
        | StringMatch(s) => 
            add_skipper( <[ 
                if (_s.Buffer.Length - _s.Position < $(s.Length : int)) false
                else {
                    if (System.String.Compare($(s : string), 0, _s.Buffer, _s.Position, $(s.Length : int)) == 0) {
                        _s.Position += $(s.Length : int);
                        true
                    } else false
                }
            ]> )
        | RegexpMatch(s) =>
            def sm = "\\G" + s;
            def re = Macros.NewSymbol();
            def mt = Macros.NewSymbol();
            t.Define(<[ decl: 
                    static $(re : name) : System.Text.RegularExpressions.Regex; 
                ]>);
            st_const.Body = <[ 
                    $(st_const.Body);
                    $(re : name) = System.Text.RegularExpressions.Regex(
                        $(sm : string), 
                        System.Text.RegularExpressions.RegexOptions.Compiled)
                ]>;
            add_skipper( <[
                def $(mt : name) = $(re : name).Match(_s.Buffer, _s.Position);
                if ($(mt : name).Success) {
                    _s.Position = $(mt : name).Index + $(mt : name).Length;
                    true
                } else false
            ]> )
        | _ => compile_expr_generic(ex)
    }
};

class TokenParserCompiler: AbstractParserCompiler 
{
    tokentype: PT.PExpr;

    public this(ctx: Typer, tokentype: PT.PExpr) {
        base(ctx, <[ ITokenParser[$tokentype] ]>);
        this.tokentype = tokentype;
    }

    protected override compile_public(main: RuleDesc): void {
        t.Define( <[ decl: public ParseGet(s: array[$tokentype]): $(main.rt : typed) * int {
                def sc = TokenScanner(s);
                def (ok, ret) = $(main.nm : name)(sc);
                if (ok) (ret, sc.Position)
                else (ret, -1)
            } ]> );
        t.Define( <[ decl: public Parse(sc: TokenScanner[$tokentype]): bool {
                def (ok, _) = $(main.nm : name)(sc);
                ok
            } ]> );
        t.Define( <[ decl: public Parse(s: array[$tokentype]): int {
                def sc = TokenScanner(s);
                def (ok, _) = $(main.nm : name)(sc);
                if (ok) sc.Position
                else -1
            } ]> );
    }
    
    protected override build_rule(rule: RuleDesc, exprs: list[PT.PExpr]): PT.ClassMember {
        <[ decl: $(rule.nm : name)(_s: TokenScanner[$tokentype]) : bool * $(rule.rt : typed) {
            def _OldPosition = _s.Position;
            def $("_current": usesite)() {
                _s.Buffer[_s.Position]
            }
            {.. $exprs}
        } ]>
    }

    protected override compile_expr(ex: Expression): PT.PExpr { 
        | MatchMatch(p) =>
            add_skipper( <[
                if (_s.Length == _s.Position) false
                else match (_s.Buffer[_s.Position]) {
                    | $p => 
                        _s.Position += 1;
                        true
                    | _ => false
                }
            ]> )
        | Sequence(MatchMatch(p),c) =>
            add_skipper( <[
                if (_s.Length == _s.Position) false
                else match (_s.Buffer[_s.Position]) {
                    | $p => 
                        _s.Position += 1;
                        $(compile_expr(c))
                    | _ => false
                }
            ]> )
        | _ => compile_expr_generic(ex)
    }
};

class Utility {
    public static flatten_tokens(t: Token, sep: Token): list[Token] {
        def separate(l) {
            if (l is []) []
            else 
                if (sep == null) l
                else sep :: l
        }
        match(t) {
            | BracesGroup(g) =>
                (Token.BeginBrace() :: flatten_tokens(g, Token.Semicolon())) + separate(Token.EndBrace() :: flatten_tokens(t.Next, sep))
            | RoundGroup(g) =>
                (Token.BeginRound() :: flatten_tokens(g, Token.Comma())) + separate(Token.EndRound() :: flatten_tokens(t.Next, sep))
            | SquareGroup(g) =>
                (Token.BeginSquare() :: flatten_tokens(g, Token.Comma())) + separate(Token.EndSquare() :: flatten_tokens(t.Next, sep))
            | LooseGroup(g) =>
                flatten_tokens(g, null) + separate(flatten_tokens(t.Next, sep))
            | QuoteGroup(_) // leave untouched: Nemerle code
            | _ =>
                if (t == null) []
                else t :: separate(flatten_tokens(t.Next, sep))
        }
    }
   
    public static p_code(t: Token): PT.PExpr.Sequence {
        MacroColorizer.PushUseSiteColor();
        try {
//            Message.Debug("p_code");
            def re = MainParser.ParseExpr(MacroColorizer.UseContext, 
                Token.QuoteGroup(t).ToString());
//            Message.Debug(re.ToString());
            def re2 = match(re) {
                | Sequence([Quoted(Expression(x))]) => x
                | Quoted(Expression(x)) => x
                | _ => throw Exc()
            }
//            Message.Debug(re2.ToString());
            match (re2) {
                | Sequence([Sequence(x)]) 
                | Sequence(x) =>
                    PT.PExpr.Sequence(x)
                | _ => 
                    PT.PExpr.Sequence([re2])
            }
        } finally {
            MacroColorizer.PopColor();
        }
    }
};

macro Parser (tp, pdef: Token)
syntax ("parser", "[", tp, "]", pdef)
{
    def pparser = parser[Nemerle.Compiler.Token] (
    matcher = <[ mutable level = 0; mutable all = [] ]> ->
        :Operator(":"):
        ( (<[ !(_current() is Nemerle.Compiler.Token.Operator(":") && level == 0 ) ]>= , .)
        & <[ all = _current() :: all ]>
          ( (:BeginBrace: | :BeginRound: | :BeginQuote: | :BeginSquare:)
            <[ level += 1 ]>
          | (:EndBrace: | :EndRound: | :EndQuote: | :EndSquare:)
            <[ level -= 1 ]>
          | .
          )
        )*
        :Operator(":"): 
        -> <[ Expression.MatchMatch(p_mcase(all.Rev())) : Expression ]>;

    charclass = <[ mutable cs = [] ]> ->
        :Operator(":"):
        ( :Identifier("white"): <[ cs = CharacterClass.Whitespace() :: cs ]>
        | :Identifier("alpha"): <[ cs = CharacterClass.Alphabetic() :: cs ]>
        | :Identifier("alnum"): <[ cs = CharacterClass.Alphanumeric() :: cs ]>
        | :Identifier("digit"): <[ cs = CharacterClass.Numeric() :: cs ]>
        | :Identifier("punct"): <[ cs = CharacterClass.Punctuation() :: cs ]>
        | :Identifier("symbl"): <[ cs = CharacterClass.Symbol() :: cs ]>
        | :Identifier("upper"): <[ cs = CharacterClass.Upper() :: cs ]>
        | :Identifier("lower"): <[ cs = CharacterClass.Lower() :: cs ]>
        )* :Operator(":"): 
        -> <[ Expression.CharClass(cs) : Expression ]>;

    term = <[ mutable ret = null; mutable ts = null ]> ->
          :Identifier("skip"): :BeginSquare: 
          (:Identifier(s): <[ ts = s ]> | :Keyword(s): <[ ts = s ]>)
          :EndSquare: :BeginRound: alt :EndRound:
          <[ ret = Expression.ChangeSkip(ts, alt) ]>
        | :Identifier(s): <[ ret = Expression.Rule(s) ]>
        | :Keyword(s): <[ ret = Expression.Rule(s) ]>
        | :QuoteGroup(g): <[ ret = Expression.Code(p_code(g)) ]>
        | :Operator("."): <[ ret = Expression.Anything() ]>
        | :BeginRound: alt :EndRound: <[ ret = alt ]>
        | <[ tokenp ]>= (:Operator(":"):)= matcher <[ ret = matcher ]>
        | <[ !tokenp ]>= (:Operator(":"):)= charclass <[ ret = charclass ]>
        | <[ !tokenp ]>= :CharLiteral(s): <[ ret = Expression.CharMatch(s) ]>
        | <[ !tokenp ]>= :StringLiteral(s): <[ ret = Expression.StringMatch(s) ]>
        | <[ !tokenp ]>= :Operator("%"): :StringLiteral(s): <[ ret = Expression.RegexpMatch(s) ]>
        -> <[ (ret) : Expression ]>;

    multi = <[ mutable ret = null ]> ->
        term <[ ret = term ]>
        ( :Operator("*"): <[ ret = Expression.Many(ret) ]>
        | :Operator("+"): <[ ret = Expression.AtLeast(1, ret) ]>
        | :Operator("!"): <[ ret = Expression.Optional(ret) ]>
        | :Operator("="): <[ ret = Expression.Epsilon(ret) ]>
        | :BeginBrace: :IntegerLiteral(n,_): :EndBrace: <[ ret = Expression.Exactly(n.val :> int, ret) ]>
        | :BeginBrace: :IntegerLiteral(n,_): :Comma: :EndBrace: <[ ret = Expression.AtLeast(n.val :> int, ret) ]>
        | :BeginBrace: :IntegerLiteral(n,_): :Comma: :IntegerLiteral(n2,_): :EndBrace: <[ ret = Expression.FromTo(n.val :> int, n2.val :> int, ret) ]>
        )!
        -> <[ (ret) : Expression ]>;

    seq = <[ mutable ret = null ]> ->
        multi <[ ret = multi ]>
        (:Comma: ! seq <[ ret = Expression.Sequence(ret, seq) ]> )! 
        -> <[ (ret) : Expression ]>;

    inter = <[ mutable ret = null ]> ->
        seq <[ ret = seq ]>
        (:Operator("&"): inter <[ ret = Expression.Intersection(ret, inter) ]> )! 
        -> <[ (ret) : Expression ]>;

    alt = <[ mutable ret = null ]> ->
        inter <[ ret = inter ]> 
        (:Operator("|"): alt <[ ret = Expression.Alternative(ret, alt) ]> )! 
        -> <[ (ret) : Expression ]>;

    rule = <[ { mutable start = <[ ]>; mutable desc = null } ]> ->
        :Identifier(n): :Operator("="): 
        (:QuoteGroup(g): :Operator("->"): <[ start = p_code(g) ]>)!
        alt
        (:Operator("->"): :QuoteGroup(h): <[ desc = match(p_code(h)) {
            | Sequence([<[ $_ : $t ]>]) =>
                RuleDesc(n, alt, start, p_code(h), ctx.MonoBindType(t)) 
            | _ => throw Exc()
        } ]>
        | <[ desc = RuleDesc(n, alt, start, <[ true ]>, ctx.MonoBindType(<[ bool ]>)) ]>)
        :Semicolon: 
        -> <[ (desc) : RuleDesc ]>; 

    pdef = <[ mutable desc = ParserDesc([],[]) ]> -> 
        (rule <[ desc = ParserDesc(rule :: desc.rules, desc.members) ]>
        | :QuoteGroup(g): <[ desc = ParserDesc(desc.rules, p_decl(g) :: desc.members) ]>
        )* 
        -> <[ (desc) : ParserDesc ]>;

    main = :BeginRound: pdef :EndRound: -> <[ (pdef) : ParserDesc ]>;
   
    <[ ctx : Typer ]>
    <[ tokenp : bool ]>

    <[ public this(ctx: Typer, tokenp: bool) {
        this.ctx = ctx;
        this.tokenp = tokenp;
    } ]>
   
    <[ p_decl(t: Token): PT.ClassMember {
        MacroColorizer.PushUseSiteColor();
        try {
//            Message.Debug("p_decl");
            def id = Token.Identifier("decl");
            id.Next = Token.Operator(":");
            id.Next.Next = t;
            def re = MainParser.ParseExpr(MacroColorizer.UseContext, 
                Token.QuoteGroup(Token.LooseGroup(id)).ToString());
//            Message.Debug(re.ToString());
            match (re) {
                | Sequence([Quoted(ClassMember(x))])
                | Quoted(ClassMember(x)) => x
                | _ => throw Exc()
            }
        } finally {
            MacroColorizer.PopColor();
        }
    } ]>

    <[ p_mcase(l: list[Token]): PT.PExpr {
        // przerobic na uzywanie cytowania i MatchCase
        MacroColorizer.PushUseSiteColor();
        try {
            def re = MainParser.ParseExpr(MacroColorizer.UseContext, string_tokens(l));
            def re2 = match(re) {
                | Sequence([x])
                | x => x
            }
            re2
        } finally {
            MacroColorizer.PopColor();
        }
    } ]>

    <[ p_code(t: Token): PT.PExpr.Sequence {
        Utility.p_code(t);
    } ]>

    <[ string_tokens(tl: SCG.IEnumerable[Token]): string {
        mutable s = "";
        foreach (t in tl) {
            s += t.ToString();
        }
        s
    } ]>
);
    
    def ctx = Nemerle.Macros.ImplicitCTX();

    def tokenp = match(tp) {
        | <[ string ]> => false
        | <[ $(_ : name) ]>
        | <[ $_ . $_ ]> => true
        | _ => throw Exc()
    }

    def ebnf = pparser(ctx, tokenp);
    def (pdef, n) = ebnf.ParseGet(Utility.flatten_tokens(pdef, null).ToArray());
    when (n == -1) throw Exc();
    def comp = if (tokenp) TokenParserCompiler(ctx, tp)
               else StringParserCompiler(ctx);
    def t = comp.compile(pdef);

    <[ $(t.ParsedName : name) ]>
}

}
